//This was created by EagleoutIce 'document creator: create_doc' using doxygen 1.8.15 and python 3.5.2
//Created: 04.05.2018 22:58:22
#ifndef EAGLE_EBOARD_HELPLIB_SERVO
    #define EAGLE_EBOARD_HELPLIB_SERVO


    


//=====================================================================================================================================================
//                                                                  Macro Definitions                                                                  
//=====================================================================================================================================================

	//import servo-class
	//shameless copy paste :D
	


	#define _useTimer1
	typedef enum { _timer1, _Nbr_16timers } timer16_Sequence_t;
	#define MIN_PULSE_WIDTH       544     // the shortest pulse sent to a servo
	#define MAX_PULSE_WIDTH      2400     // the longest pulse sent to a servo
	#define DEFAULT_PULSE_WIDTH  1500     // default pulse width when servo is attached
	#define REFRESH_INTERVAL    20000     // minumim time to refresh servos in microseconds

	#define SERVOS_PER_TIMER       12     // the maximum number of servos controlled by one timer
	#define MAX_SERVOS   (_Nbr_16timers  * SERVOS_PER_TIMER)

	#define INVALID_SERVO         255     // flag indicating an invalid servo index

//=====================================================================================================================================================
//                                                                     ServoPin_t                                                                      
//=====================================================================================================================================================

	

	typedef struct  {
		

	  uint8_t nbr        :6 ;             // a pin number from 0 to 63
		

	  uint8_t isActive   :1 ;             // true if this channel is enabled, pin not pulsed if false
	} ServoPin_t;

//=====================================================================================================================================================
//                                                                       Servo_t                                                                       
//=====================================================================================================================================================

	

	typedef struct {
		

	  ServoPin_t Pin;
		

	  volatile unsigned int ticks;
	} servo_t;

//=====================================================================================================================================================
//                                                                        Servo                                                                        
//=====================================================================================================================================================

    //-------------------------------------------------------------------------------------------------------------------------------------------------
    //                                                                    class                                                                        
    //-------------------------------------------------------------------------------------------------------------------------------------------------

	


	

	class Servo {
	  public:
			

	    Servo();
        

        uint8_t attach(int pin);           // attach the given pin to the next free channel, sets pinMode, returns channel number or 0 if failure
        

        uint8_t attach(int pin, int min, int max); // as above but also sets min and max values for writes.
        

        void detach();
        

        void write(int value);             // if value is < 200 its treated as an angle, otherwise as pulse width in microseconds
        

        void writeMicroseconds(int value); // Write pulse width in microseconds
        

        inline int read();                        // returns current pulse width as an angle between 0 and 180 degrees
        

        int readMicroseconds();            // returns current pulse width in microseconds for this servo (was read_us() in first release)
        

        inline bool attached();                   // return true if this servo is attached, otherwise false
	  private:
			 

	     uint8_t servoIndex;               // index into the channel data for this servo
			 

	     int8_t min;                       // minimum is this value times 4 added to MIN_PULSE_WIDTH
			 

	     int8_t max;                       // maximum is this value times 4 added to MAX_PULSE_WIDTH
	};
	


    //-------------------------------------------------------------------------------------------------------------------------------------------------
    //                                                                definitions                                                                      
    //-------------------------------------------------------------------------------------------------------------------------------------------------

	#define usToTicks(_us)    (( clockCyclesPerMicrosecond()* _us) / 8)     // converts microseconds to tick (assumes prescale of 8)  // 12 Aug 2009
	#define ticksToUs(_ticks) (( (unsigned)_ticks * 8)/ clockCyclesPerMicrosecond() ) // converts from ticks back to microseconds


	#define TRIM_DURATION       2                               // compensation ticks to trim adjust for digitalWrite delays // 12 August 2009

	//#define NBR_TIMERS        (MAX_SERVOS / SERVOS_PER_TIMER)

	static servo_t servos[MAX_SERVOS];                          // static array of servo structures
	static volatile int8_t Channel[_Nbr_16timers ];             // counter for the servo being pulsed for each timer (or -1 if refresh interval)

	uint8_t ServoCount = 0;                                     // the total number of attached servos

	// convenience macros
	#define SERVO_INDEX_TO_TIMER(_servo_nbr) ((timer16_Sequence_t)(_servo_nbr / SERVOS_PER_TIMER)) // returns the timer controlling this servo
	#define SERVO_INDEX_TO_CHANNEL(_servo_nbr) (_servo_nbr % SERVOS_PER_TIMER)       // returns the index of the servo on this timer
	#define SERVO_INDEX(_timer,_channel)  ((_timer*SERVOS_PER_TIMER) + _channel)     // macro to access servo index by timer and channel
	#define SERVO(_timer,_channel)  (servos[SERVO_INDEX(_timer,_channel)])            // macro to access servo class by timer and channel

	#define SERVO_MIN() (MIN_PULSE_WIDTH - this->min * 4)  // minimum value in uS for this servo
	#define SERVO_MAX() (MAX_PULSE_WIDTH - this->max * 4)  // maximum value in uS for this servo

	static inline void handle_interrupts(timer16_Sequence_t timer, volatile uint16_t *TCNTn, volatile uint16_t* OCRnA)
	{
	  if( Channel[timer] < 0 )
	    *TCNTn = 0; // channel set to -1 indicated that refresh interval completed so reset the timer
	  else{
	    if( SERVO_INDEX(timer,Channel[timer]) < ServoCount && SERVO(timer,Channel[timer]).Pin.isActive == true )
	      digitalWrite( SERVO(timer,Channel[timer]).Pin.nbr,LOW); // pulse this channel low if activated
	  }

	  Channel[timer]++;    // increment to the next channel
	  if( SERVO_INDEX(timer,Channel[timer]) < ServoCount && Channel[timer] < SERVOS_PER_TIMER) {
	    *OCRnA = *TCNTn + SERVO(timer,Channel[timer]).ticks;
	    if(SERVO(timer,Channel[timer]).Pin.isActive == true)     // check if activated
	      digitalWrite( SERVO(timer,Channel[timer]).Pin.nbr,HIGH); // its an active channel so pulse it high
	  }
	  else {
	    // finished all channels so wait for the refresh period to expire before starting over
	    if( ((unsigned)*TCNTn) + 4 < usToTicks(REFRESH_INTERVAL) )  // allow a few ticks to ensure the next OCR1A not missed
	      *OCRnA = (unsigned int)usToTicks(REFRESH_INTERVAL);
	    else
	      *OCRnA = *TCNTn + 4;  // at least REFRESH_INTERVAL has elapsed
	    Channel[timer] = -1; // this will get incremented at the end of the refresh period to start again at the first channel
	  }
	}

	#ifndef WIRING // Wiring pre-defines signal handlers so don't define any if compiling for the Wiring platform
	// Interrupt handlers for Arduino
	  #if defined(_useTimer1)
	  SIGNAL (TIMER1_COMPA_vect)
	  {
	    handle_interrupts(_timer1, &TCNT1, &OCR1A);
	  }
	  #endif
	#elif defined WIRING
	  // Interrupt handlers for Wiring
	  #if defined(_useTimer1)
	  void Timer1Service()
	  {
	    handle_interrupts(_timer1, &TCNT1, &OCR1A);
	  }
	  #endif
	#endif

	inline int Servo::read() // return the value as degrees
	{
	  return  map( this->readMicroseconds()+1, SERVO_MIN(), SERVO_MAX(), 0, 180);
	}


Servo::Servo() {
	  if( ServoCount < MAX_SERVOS) {
	    this->servoIndex = ServoCount++;                    // assign a servo index to this instance
	    servos[this->servoIndex].ticks = usToTicks(DEFAULT_PULSE_WIDTH);   // store default values  - 12 Aug 2009
	  }
	  else
	    this->servoIndex = INVALID_SERVO ;  // too many servos
	}

	uint8_t Servo::attach(int pin) {
	  return this->attach(pin, MIN_PULSE_WIDTH, MAX_PULSE_WIDTH);
	}

	uint8_t Servo::attach(int pin, int min, int max) {
	  if(this->servoIndex < MAX_SERVOS ) {
	    pinMode( pin, OUTPUT) ;                                   // set servo pin to output
	    servos[this->servoIndex].Pin.nbr = pin;
	    // todo min/max check: abs(min - MIN_PULSE_WIDTH) /4 < 128
	    this->min  = (MIN_PULSE_WIDTH - min)/4; //resolution of min/max is 4 uS
	    this->max  = (MAX_PULSE_WIDTH - max)/4;
	    // initialize the timer if it has not already been initialized
	    timer16_Sequence_t timer = SERVO_INDEX_TO_TIMER(servoIndex);
	    if(isTimerActive(timer) == false)
	      initISR(timer);
	    servos[this->servoIndex].Pin.isActive = true;  // this must be set after the check for isTimerActive
	  }
	  return this->servoIndex ;
	}

	void Servo::detach() {
	  servos[this->servoIndex].Pin.isActive = false;
	  timer16_Sequence_t timer = SERVO_INDEX_TO_TIMER(servoIndex);
	  if(isTimerActive(timer) == false) {
	    finISR(timer);
	  }
	}

	void Servo::write(int value) {
	  if(value < MIN_PULSE_WIDTH)
	  {  // treat values less than 544 as angles in degrees (valid values in microseconds are handled as microseconds)
	    if(value < 0) value = 0;
	    if(value > 180) value = 180;
	    value = map(value, 0, 180, SERVO_MIN(),  SERVO_MAX());
	  }
	  this->writeMicroseconds(value);
	}

	void Servo::writeMicroseconds(int value) {
	  // calculate and store the values for the given channel
	  byte channel = this->servoIndex;
	  if( (channel < MAX_SERVOS) )   // ensure channel is valid
	  {
	    if( value < SERVO_MIN() )          // ensure pulse width is valid
	      value = SERVO_MIN();
	    else if( value > SERVO_MAX() )
	      value = SERVO_MAX();

	    value = value - TRIM_DURATION;
	    value = usToTicks(value);  // convert to ticks after compensating for interrupt overhead - 12 Aug 2009

	    uint8_t oldSREG = SREG;
	    cli();
	    servos[channel].ticks = value;
	    SREG = oldSREG;
	  }
	}


	int Servo::readMicroseconds() {
	  unsigned int pulsewidth;
	  if( this->servoIndex != INVALID_SERVO )
	    pulsewidth = ticksToUs(servos[this->servoIndex].ticks)  + TRIM_DURATION ;   // 12 aug 2009
	  else
	    pulsewidth  = 0;

	  return pulsewidth;
	}

	inline bool Servo::attached() {
	  return servos[this->servoIndex].Pin.isActive ;
	}

	//========= TIMER 
	
		static void initISR(timer16_Sequence_t timer) {
	  if(timer == _timer1) {
	    TCCR1A = 0;             // normal counting mode
	    TCCR1B = _BV(CS11);     // set prescaler of 8
	    TCNT1 = 0;              // clear the timer count
	    TIFR1 |= _BV(OCF1A);     // clear any pending interrupts;
	    TIMSK1 |=  _BV(OCIE1A) ; // enable the output compare interrupt
	    #if defined(WIRING)
	      timerAttach(TIMER1OUTCOMPAREA_INT, Timer1Service);
	    #endif
	    }
	}

	static void finISR(timer16_Sequence_t timer) {
	  #if defined WIRING   // Wiring
	  if(timer == _timer1) {
	    TIMSK &=  ~_BV(OCIE1A) ;  // disable timer 1 output compare interrupt
	    timerDetach(TIMER1OUTCOMPAREA_INT);
	  }
	  else if(timer == _timer3) {
	    ETIMSK &= ~_BV(OCIE3A);    // disable the timer3 output compare A interrupt
	    timerDetach(TIMER3OUTCOMPAREA_INT);
	  }
	  #else
	  //For arduino - in future: call here to a currently undefined function to reset the timer
	  (void) timer;  // squash "unused parameter 'timer' [-Wunused-parameter]" warning
	  #endif
	}
    
    static bool isTimerActive(timer16_Sequence_t timer) {
	  // returns true if any servo is active on this timer
	  for(uint8_t channel=0; channel < SERVOS_PER_TIMER; channel++) {
	    if(SERVO(timer,channel).Pin.isActive == true)
	      return true;
	  }
	  return false;
	}


	

	

#endif
